"""Tests for RetryStreamResponse and AsyncRetryStreamResponse."""

import pytest
from pydantic import BaseModel

from mirascope import llm

from .conftest import CONNECTION_ERROR, RATE_LIMIT_ERROR, SERVER_ERROR, TIMEOUT_ERROR
from .mock_provider import MockProvider

# --- Sync streaming tests ---


def test_stream_succeeds_first_attempt(mock_provider: MockProvider) -> None:
    """Test that a successful stream completes without raising StreamRestarted."""

    @llm.retry(max_retries=2)
    @llm.call("mock/test-model")
    def greet() -> str:
        return "Hello"

    response = greet.stream()

    # Consume the stream
    chunks = list(response.text_stream())

    assert "mock " in "".join(chunks)
    assert "response" in "".join(chunks)
    assert len(response.retry_failures) == 0
    assert mock_provider.stream_count == 1


def test_stream_raises_stream_restarted_on_error(mock_provider: MockProvider) -> None:
    """Test that StreamRestarted is raised when a retryable error occurs mid-stream."""
    mock_provider.set_stream_exceptions([CONNECTION_ERROR])

    @llm.retry(max_retries=2)
    @llm.call("mock/test-model")
    def greet() -> str:
        return "Hello"

    response = greet.stream()

    # First iteration should raise StreamRestarted
    with pytest.raises(llm.StreamRestarted) as exc_info:
        list(response.text_stream())

    assert exc_info.value.failure.exception is CONNECTION_ERROR
    assert len(response.retry_failures) == 1
    assert [f.exception for f in response.retry_failures] == [CONNECTION_ERROR]


def test_stream_can_continue_after_restart(mock_provider: MockProvider) -> None:
    """Test that user can re-iterate after catching StreamRestarted."""
    mock_provider.set_stream_exceptions([CONNECTION_ERROR])

    @llm.retry(max_retries=2)
    @llm.call("mock/test-model")
    def greet() -> str:
        return "Hello"

    response = greet.stream()
    restart_count = 0

    # Use the recommended pattern: catch StreamRestarted and re-iterate
    while True:
        try:
            chunks = list(response.text_stream())
            break
        except llm.StreamRestarted:
            restart_count += 1

    assert restart_count == 1
    assert "mock " in "".join(chunks)
    assert "response" in "".join(chunks)
    assert len(response.retry_failures) == 1
    assert mock_provider.stream_count == 2


def test_stream_raises_retries_exhausted_after_max_attempts(
    mock_provider: MockProvider,
) -> None:
    """Test that RetriesExhausted is raised when max attempts exhausted."""
    # Set more errors than max_attempts allows
    mock_provider.set_stream_exceptions([CONNECTION_ERROR, RATE_LIMIT_ERROR])

    @llm.retry(max_retries=1)
    @llm.call("mock/test-model")
    def greet() -> str:
        return "Hello"

    response = greet.stream()

    # First attempt fails, catch StreamRestarted
    with pytest.raises(llm.StreamRestarted):
        list(response.text_stream())

    # Second attempt also fails, should raise RetriesExhausted (not StreamRestarted)
    with pytest.raises(llm.RetriesExhausted) as exc_info:
        list(response.text_stream())

    # RetriesExhausted contains both failures
    assert len(exc_info.value.failures) == 2
    assert isinstance(exc_info.value.failures[0].exception, llm.ConnectionError)
    assert isinstance(exc_info.value.failures[1].exception, llm.RateLimitError)
    assert mock_provider.stream_count == 2


def test_stream_retry_failures_accessible(mock_provider: MockProvider) -> None:
    """Test that retry_failures is accessible from the response."""
    mock_provider.set_stream_exceptions([SERVER_ERROR])

    @llm.retry(max_retries=4)
    @llm.call("mock/test-model")
    def greet() -> str:
        return "Hello"

    response = greet.stream()

    # Initial state
    assert response.model.retry_config.max_retries == 4
    assert len(response.retry_failures) == 0

    # After first error
    with pytest.raises(llm.StreamRestarted):
        list(response.text_stream())

    assert len(response.retry_failures) == 1


def test_stream_model_property(mock_provider: MockProvider) -> None:  # noqa: ARG001
    """Test that model property returns RetryModel."""

    @llm.retry(max_retries=2)
    @llm.call("mock/test-model")
    def greet() -> str:
        return "Hello"

    response = greet.stream()

    assert isinstance(response.model, llm.RetryModel)
    assert response.model.model_id == "mock/test-model"


def test_stream_retry_config_property(mock_provider: MockProvider) -> None:  # noqa: ARG001
    """Test that retry_config property returns the retry configuration."""

    @llm.retry(max_retries=3)
    @llm.call("mock/test-model")
    def greet() -> str:
        return "Hello"

    response = greet.stream()

    assert response.retry_config.max_retries == 3


@pytest.mark.asyncio
async def test_async_stream_retry_config_property(
    mock_provider: MockProvider,  # noqa: ARG001
) -> None:
    """Test that async retry_config property returns the retry configuration."""

    @llm.retry(max_retries=3)
    @llm.call("mock/test-model")
    async def greet() -> str:
        return "Hello"

    response = await greet.stream()

    assert response.retry_config.max_retries == 3


# --- Async streaming tests ---


@pytest.mark.asyncio
async def test_stream_async_succeeds_first_attempt(mock_provider: MockProvider) -> None:
    """Test that a successful async stream completes without raising StreamRestarted."""

    @llm.retry(max_retries=2)
    @llm.call("mock/test-model")
    async def greet() -> str:
        return "Hello"

    response = await greet.stream()

    # Consume the stream
    chunks = [chunk async for chunk in response.text_stream()]

    assert "mock " in "".join(chunks)
    assert "response" in "".join(chunks)
    assert len(response.retry_failures) == 0
    assert mock_provider.stream_count == 1


@pytest.mark.asyncio
async def test_stream_async_raises_stream_restarted_on_error(
    mock_provider: MockProvider,
) -> None:
    """Test that StreamRestarted is raised when a retryable error occurs mid-stream."""
    mock_provider.set_stream_exceptions([TIMEOUT_ERROR])

    @llm.retry(max_retries=2)
    @llm.call("mock/test-model")
    async def greet() -> str:
        return "Hello"

    response = await greet.stream()

    # First iteration should raise StreamRestarted
    with pytest.raises(llm.StreamRestarted) as exc_info:
        _ = [chunk async for chunk in response.text_stream()]

    assert exc_info.value.failure.exception is TIMEOUT_ERROR
    assert len(response.retry_failures) == 1


@pytest.mark.asyncio
async def test_stream_async_can_continue_after_restart(
    mock_provider: MockProvider,
) -> None:
    """Test that user can re-iterate after catching StreamRestarted."""
    mock_provider.set_stream_exceptions([SERVER_ERROR])

    @llm.retry(max_retries=2)
    @llm.call("mock/test-model")
    async def greet() -> str:
        return "Hello"

    response = await greet.stream()
    restart_count = 0

    # Use the recommended pattern: catch StreamRestarted and re-iterate
    while True:
        try:
            chunks = [chunk async for chunk in response.text_stream()]
            break
        except llm.StreamRestarted:
            restart_count += 1

    assert restart_count == 1
    assert "mock " in "".join(chunks)
    assert "response" in "".join(chunks)
    assert len(response.retry_failures) == 1
    assert mock_provider.stream_count == 2


@pytest.mark.asyncio
async def test_stream_async_raises_retries_exhausted_after_max_attempts(
    mock_provider: MockProvider,
) -> None:
    """Test that RetriesExhausted is raised when max attempts exhausted."""
    mock_provider.set_stream_exceptions([CONNECTION_ERROR, SERVER_ERROR])

    @llm.retry(max_retries=1)
    @llm.call("mock/test-model")
    async def greet() -> str:
        return "Hello"

    response = await greet.stream()

    # First attempt fails, catch StreamRestarted
    with pytest.raises(llm.StreamRestarted):
        _ = [chunk async for chunk in response.text_stream()]

    # Second attempt also fails, should raise RetriesExhausted
    with pytest.raises(llm.RetriesExhausted) as exc_info:
        _ = [chunk async for chunk in response.text_stream()]

    # RetriesExhausted contains both failures
    assert len(exc_info.value.failures) == 2
    assert isinstance(exc_info.value.failures[0].exception, llm.ConnectionError)
    assert isinstance(exc_info.value.failures[1].exception, llm.ServerError)
    assert mock_provider.stream_count == 2


# --- Sync RetryStreamResponse property tests ---


def test_delegated_properties(mock_provider: MockProvider) -> None:
    """Test that all properties delegate to the wrapped stream response."""

    @llm.retry(max_retries=1)
    @llm.call("mock/test-model")
    def greet() -> str:
        return "Hello"

    response = greet.stream()

    # Consume the stream to populate properties
    list(response.chunk_stream())

    # Test all delegated properties
    assert response.raw_stream_events is not None
    assert response.chunks is not None
    assert response.provider_id == "mock"
    assert response.model_id == "mock/test-model"
    assert response.provider_model_name == "test-model"
    assert response.params is not None
    assert response.toolkit is not None
    assert response.messages is not None
    assert response.content is not None
    assert response.texts is not None
    assert response.tool_calls is not None
    assert response.thoughts is not None
    _ = response.finish_reason
    _ = response.usage
    assert response.consumed is True
    assert response.format is None
    assert response.parse() is None


def test_finish_method(mock_provider: MockProvider) -> None:
    """Test that finish() consumes the stream."""

    @llm.retry(max_retries=1)
    @llm.call("mock/test-model")
    def greet() -> str:
        return "Hello"

    response = greet.stream()
    assert response.consumed is False

    response.finish()

    assert response.consumed is True
    assert mock_provider.stream_count == 1


def test_pretty_stream(mock_provider: MockProvider) -> None:
    """Test that pretty_stream() yields formatted text."""

    @llm.retry(max_retries=1)
    @llm.call("mock/test-model")
    def greet() -> str:
        return "Hello"

    response = greet.stream()

    output = list(response.pretty_stream())

    # Should have content from the mock stream
    full_output = "".join(output)
    assert "mock " in full_output or "response" in full_output
    assert mock_provider.stream_count == 1


def test_streams_method(mock_provider: MockProvider) -> None:
    """Test that streams() yields Stream objects."""

    @llm.retry(max_retries=1)
    @llm.call("mock/test-model")
    def greet() -> str:
        return "Hello"

    response = greet.stream()
    streams = list(response.streams())

    assert len(streams) == 1
    assert streams[0].content_type == "text"
    assert mock_provider.stream_count == 1


def test_execute_tools_sync(mock_provider: MockProvider) -> None:  # noqa: ARG001
    """Test that execute_tools delegates to wrapped response."""

    @llm.retry(max_retries=1)
    @llm.call("mock/test-model")
    def greet() -> str:
        return "Hello"

    response = greet.stream()
    response.finish()

    result = response.execute_tools()
    assert result == []


def test_pretty_method(mock_provider: MockProvider) -> None:
    """Test that pretty() returns formatted text after consumption."""

    @llm.retry(max_retries=1)
    @llm.call("mock/test-model")
    def greet() -> str:
        return "Hello"

    response = greet.stream()
    response.finish()

    output = response.pretty()
    assert "mock response" in output


# --- Async RetryStreamResponse property tests ---


@pytest.mark.asyncio
async def test_async_delegated_properties(mock_provider: MockProvider) -> None:
    """Test that all properties delegate to the wrapped async stream response."""

    @llm.retry(max_retries=1)
    @llm.call("mock/test-model")
    async def greet() -> str:
        return "Hello"

    response = await greet.stream()

    # Consume the stream to populate properties
    async for _ in response.chunk_stream():
        pass

    # Test all delegated properties
    assert response.raw_stream_events is not None
    assert response.chunks is not None
    assert response.provider_id == "mock"
    assert response.model_id == "mock/test-model"
    assert response.provider_model_name == "test-model"
    assert response.params is not None
    assert response.toolkit is not None
    assert response.messages is not None
    assert response.content is not None
    assert response.texts is not None
    assert response.tool_calls is not None
    assert response.thoughts is not None
    _ = response.finish_reason
    _ = response.usage
    assert response.consumed is True
    assert response.format is None
    assert response.parse() is None


@pytest.mark.asyncio
async def test_async_finish_method(mock_provider: MockProvider) -> None:
    """Test that finish() consumes the async stream."""

    @llm.retry(max_retries=1)
    @llm.call("mock/test-model")
    async def greet() -> str:
        return "Hello"

    response = await greet.stream()
    assert response.consumed is False

    await response.finish()

    assert response.consumed is True
    assert mock_provider.stream_count == 1


@pytest.mark.asyncio
async def test_async_pretty_stream(mock_provider: MockProvider) -> None:
    """Test that pretty_stream() yields formatted text."""

    @llm.retry(max_retries=1)
    @llm.call("mock/test-model")
    async def greet() -> str:
        return "Hello"

    response = await greet.stream()

    output = [chunk async for chunk in response.pretty_stream()]

    # Should have content from the mock stream
    full_output = "".join(output)
    assert "mock " in full_output or "response" in full_output
    assert mock_provider.stream_count == 1


@pytest.mark.asyncio
async def test_async_streams_method(mock_provider: MockProvider) -> None:
    """Test that streams() yields AsyncStream objects."""

    @llm.retry(max_retries=1)
    @llm.call("mock/test-model")
    async def greet() -> str:
        return "Hello"

    response = await greet.stream()
    streams = [s async for s in response.streams()]

    assert len(streams) == 1
    assert streams[0].content_type == "text"
    assert mock_provider.stream_count == 1


@pytest.mark.asyncio
async def test_async_execute_tools(mock_provider: MockProvider) -> None:  # noqa: ARG001
    """Test that execute_tools delegates to wrapped response."""

    @llm.retry(max_retries=1)
    @llm.call("mock/test-model")
    async def greet() -> str:
        return "Hello"

    response = await greet.stream()
    await response.finish()

    result = await response.execute_tools()
    assert result == []


@pytest.mark.asyncio
async def test_async_pretty_method(mock_provider: MockProvider) -> None:
    """Test that pretty() returns formatted text after consumption."""

    @llm.retry(max_retries=1)
    @llm.call("mock/test-model")
    async def greet() -> str:
        return "Hello"

    response = await greet.stream()
    await response.finish()

    output = response.pretty()
    assert "mock response" in output


# --- Sync structured_stream tests ---


def test_structured_stream_raises_without_format(
    mock_provider: MockProvider,  # noqa: ARG001
) -> None:
    """Test that structured_stream raises ValueError without format."""

    @llm.retry(max_retries=1)
    @llm.call("mock/test-model")
    def greet() -> str:
        return "Hello"

    response = greet.stream()

    with pytest.raises(ValueError, match="structured_stream\\(\\) requires format"):
        list(response.structured_stream())


def test_structured_stream_yields_partials(mock_provider: MockProvider) -> None:
    """Test that structured_stream yields partial outputs."""

    class Person(BaseModel):
        name: str
        age: int

    # Set up mock to return valid JSON
    mock_provider.set_stream_text('{"name": "Alice", "age": 30}')

    retry_model = llm.retry_model("mock/test-model", max_retries=1)

    response = retry_model.stream("Get person", format=Person)

    partials = list(response.structured_stream())

    # Should have yielded at least one partial
    assert len(partials) >= 1


def test_structured_stream_raises_for_output_parser(
    mock_provider: MockProvider,  # noqa: ARG001
) -> None:
    """Test that structured_stream raises NotImplementedError for OutputParser."""

    @llm.output_parser(formatting_instructions="Return text")
    def custom_parser(response: llm.RootResponse[llm.Toolkit, None]) -> str:
        return response.text()

    retry_model = llm.retry_model("mock/test-model", max_retries=1)

    response = retry_model.stream("Hello", format=custom_parser)

    with pytest.raises(
        NotImplementedError, match="structured_stream\\(\\) not supported"
    ):
        list(response.structured_stream())


# --- Async structured_stream tests ---


@pytest.mark.asyncio
async def test_async_structured_stream_raises_without_format(
    mock_provider: MockProvider,  # noqa: ARG001
) -> None:
    """Test that structured_stream raises ValueError without format."""

    @llm.retry(max_retries=1)
    @llm.call("mock/test-model")
    async def greet() -> str:
        return "Hello"

    response = await greet.stream()

    with pytest.raises(ValueError, match="structured_stream\\(\\) requires format"):
        _ = [chunk async for chunk in response.structured_stream()]


@pytest.mark.asyncio
async def test_async_structured_stream_yields_partials(
    mock_provider: MockProvider,
) -> None:
    """Test that async structured_stream yields partial outputs."""

    class Person(BaseModel):
        name: str
        age: int

    # Set up mock to return valid JSON
    mock_provider.set_stream_text('{"name": "Bob", "age": 25}')

    retry_model = llm.retry_model("mock/test-model", max_retries=1)

    response = await retry_model.stream_async("Get person", format=Person)

    partials = [p async for p in response.structured_stream()]

    # Should have yielded at least one partial
    assert len(partials) >= 1


@pytest.mark.asyncio
async def test_async_structured_stream_raises_for_output_parser(
    mock_provider: MockProvider,  # noqa: ARG001
) -> None:
    """Test that async structured_stream raises NotImplementedError for OutputParser."""

    @llm.output_parser(formatting_instructions="Return text")
    def custom_parser(response: llm.RootResponse[llm.Toolkit, None]) -> str:
        return response.text()

    retry_model = llm.retry_model("mock/test-model", max_retries=1)

    response = await retry_model.stream_async("Hello", format=custom_parser)

    with pytest.raises(
        NotImplementedError, match="structured_stream\\(\\) not supported"
    ):
        _ = [p async for p in response.structured_stream()]


# --- Sync resume tests ---


def test_resume_returns_retry_stream_response(mock_provider: MockProvider) -> None:
    """Test that resume returns a new RetryStreamResponse."""

    @llm.retry(max_retries=2)
    @llm.call("mock/test-model")
    def greet() -> str:
        return "Hello"

    response = greet.stream()
    response.finish()

    resumed = response.resume("Follow up")

    assert isinstance(resumed, llm.RetryStreamResponse)
    assert mock_provider.stream_count == 2


# --- Async resume tests ---


@pytest.mark.asyncio
async def test_async_resume_returns_async_retry_stream_response(
    mock_provider: MockProvider,
) -> None:
    """Test that resume returns a new AsyncRetryStreamResponse."""

    @llm.retry(max_retries=2)
    @llm.call("mock/test-model")
    async def greet() -> str:
        return "Hello"

    response = await greet.stream()
    await response.finish()

    resumed = await response.resume("Follow up")

    assert isinstance(resumed, llm.AsyncRetryStreamResponse)
    assert mock_provider.stream_count == 2


@pytest.mark.asyncio
async def test_async_resume_stream_factory_called_on_retry(
    mock_provider: MockProvider,
) -> None:
    """Test that resume's stream_factory is invoked when a retry happens."""

    @llm.retry(max_retries=2)
    @llm.call("mock/test-model")
    async def greet() -> str:
        return "Hello"

    response = await greet.stream()
    await response.finish()

    # Set up error for resumed stream
    mock_provider.set_stream_exceptions([CONNECTION_ERROR])

    resumed = await response.resume("Follow up")

    # First iteration fails, raises StreamRestarted
    with pytest.raises(llm.StreamRestarted):
        _ = [chunk async for chunk in resumed.text_stream()]

    # Stream factory was called for the retry:
    # 1 = original stream, 2 = resume initial stream, 3 = retry (via stream_factory)
    assert mock_provider.stream_count == 3

    # Can successfully complete after retry (uses stream #3 that was already created)
    chunks = [chunk async for chunk in resumed.text_stream()]
    assert "mock " in "".join(chunks) or "response" in "".join(chunks)
    # No new stream created - we're using the one created by stream_factory
    assert mock_provider.stream_count == 3


# --- Sync resume with fallback tests ---


def test_resume_uses_fallback_model_when_original_succeeded_on_fallback(
    mock_provider: MockProvider,
) -> None:
    """Test that stream resume uses the fallback model when original response succeeded on fallback."""
    # First stream: primary fails, fallback succeeds
    mock_provider.set_stream_exceptions([CONNECTION_ERROR])

    retry_model = llm.retry_model(
        "mock/primary", max_retries=0, fallback_models=["mock/fallback"]
    )

    response = retry_model.stream("Hello")

    # First iteration fails (primary), raises StreamRestarted
    with pytest.raises(llm.StreamRestarted):
        list(response.chunk_stream())

    # After restart, should be on fallback model
    assert response.model.model_id == "mock/fallback"

    # Second iteration succeeds on fallback
    list(response.chunk_stream())

    # Resume should use the fallback model (which succeeded)
    resumed = response.resume("Follow up")

    # Should start on fallback (no errors set)
    assert resumed.model.model_id == "mock/fallback"

    # Should succeed on fallback
    chunks = list(resumed.chunk_stream())
    assert len(chunks) > 0


def test_resume_can_fall_back_to_original_if_fallback_fails(
    mock_provider: MockProvider,
) -> None:
    """Test that stream resume can fall back to original model if the fallback model fails."""
    # First stream: primary fails, fallback succeeds
    mock_provider.set_stream_exceptions([CONNECTION_ERROR])

    retry_model = llm.retry_model(
        "mock/primary", max_retries=0, fallback_models=["mock/fallback"]
    )

    response = retry_model.stream("Hello")

    # First iteration fails (primary), raises StreamRestarted
    with pytest.raises(llm.StreamRestarted):
        list(response.chunk_stream())

    # After restart, should be on fallback model
    assert response.model.model_id == "mock/fallback"

    # Second iteration succeeds on fallback
    list(response.chunk_stream())

    # For resume: fallback fails, primary succeeds
    mock_provider.set_stream_exceptions([RATE_LIMIT_ERROR])

    resumed = response.resume("Follow up")

    # Should start on fallback model (which succeeded before)
    assert resumed.model.model_id == "mock/fallback"

    # First iteration of resume fails (fallback), raises StreamRestarted
    with pytest.raises(llm.StreamRestarted):
        list(resumed.chunk_stream())

    # After restart, should be on primary model
    assert resumed.model.model_id == "mock/primary"
    assert len(resumed.retry_failures) == 1
    assert resumed.retry_failures[0].model.model_id == "mock/fallback"

    # Second iteration succeeds on primary
    chunks = list(resumed.chunk_stream())
    assert len(chunks) > 0


# --- Async resume with fallback tests ---


@pytest.mark.asyncio
async def test_async_resume_uses_fallback_model_when_original_succeeded_on_fallback(
    mock_provider: MockProvider,
) -> None:
    """Test that async stream resume uses the fallback model when original response succeeded on fallback."""
    # First stream: primary fails, fallback succeeds
    mock_provider.set_stream_exceptions([CONNECTION_ERROR])

    retry_model = llm.retry_model(
        "mock/primary", max_retries=0, fallback_models=["mock/fallback"]
    )

    response = await retry_model.stream_async("Hello")

    # First iteration fails (primary), raises StreamRestarted
    with pytest.raises(llm.StreamRestarted):
        _ = [chunk async for chunk in response.chunk_stream()]

    # After restart, should be on fallback model
    assert response.model.model_id == "mock/fallback"

    # Second iteration succeeds on fallback
    _ = [chunk async for chunk in response.chunk_stream()]

    # Resume should use the fallback model (which succeeded)
    resumed = await response.resume("Follow up")

    # Should start on fallback (no errors set)
    assert resumed.model.model_id == "mock/fallback"

    # Should succeed on fallback
    chunks = [chunk async for chunk in resumed.chunk_stream()]
    assert len(chunks) > 0


@pytest.mark.asyncio
async def test_async_resume_can_fall_back_to_original_if_fallback_fails(
    mock_provider: MockProvider,
) -> None:
    """Test that async stream resume can fall back to original model if the fallback model fails."""
    # First stream: primary fails, fallback succeeds
    mock_provider.set_stream_exceptions([CONNECTION_ERROR])

    retry_model = llm.retry_model(
        "mock/primary", max_retries=0, fallback_models=["mock/fallback"]
    )

    response = await retry_model.stream_async("Hello")

    # First iteration fails (primary), raises StreamRestarted
    with pytest.raises(llm.StreamRestarted):
        _ = [chunk async for chunk in response.chunk_stream()]

    # After restart, should be on fallback model
    assert response.model.model_id == "mock/fallback"

    # Second iteration succeeds on fallback
    _ = [chunk async for chunk in response.chunk_stream()]

    # For resume: fallback fails, primary succeeds
    mock_provider.set_stream_exceptions([RATE_LIMIT_ERROR])

    resumed = await response.resume("Follow up")

    # Should start on fallback model (which succeeded before)
    assert resumed.model.model_id == "mock/fallback"

    # First iteration of resume fails (fallback), raises StreamRestarted
    with pytest.raises(llm.StreamRestarted):
        _ = [chunk async for chunk in resumed.chunk_stream()]

    # After restart, should be on primary model
    assert resumed.model.model_id == "mock/primary"
    assert len(resumed.retry_failures) == 1
    assert resumed.retry_failures[0].model.model_id == "mock/fallback"

    # Second iteration succeeds on primary
    chunks = [chunk async for chunk in resumed.chunk_stream()]
    assert len(chunks) > 0


# --- Context stream retry_config tests ---


def test_context_stream_retry_config_property(
    mock_provider: MockProvider,  # noqa: ARG001
) -> None:
    """Test that ContextRetryStreamResponse.retry_config property returns the retry configuration."""
    retry_model = llm.retry_model("mock/test-model", max_retries=3)
    ctx = llm.Context(deps=None)

    response = retry_model.context_stream("Hello", ctx=ctx)

    assert response.retry_config.max_retries == 3


@pytest.mark.asyncio
async def test_async_context_stream_retry_config_property(
    mock_provider: MockProvider,  # noqa: ARG001
) -> None:
    """Test that AsyncContextRetryStreamResponse.retry_config property returns the retry configuration."""
    retry_model = llm.retry_model("mock/test-model", max_retries=3)
    ctx = llm.Context(deps=None)

    response = await retry_model.context_stream_async("Hello", ctx=ctx)

    assert response.retry_config.max_retries == 3



from inline_snapshot import snapshot

test_snapshot = snapshot(
    {
        "response1": "General Kenobi! This is where the fun begins. How may I assist you in this galaxy far, far away today?",
        "usage1": {
            "input_tokens": 3698,
            "output_tokens": 26,
            "cache_read_tokens": 0,
            "cache_write_tokens": 0,
            "reasoning_tokens": 0,
            "provider_tool_usage": None,
            "raw": "CompletionUsage(completion_tokens=26, prompt_tokens=3698, total_tokens=3724, completion_tokens_details=CompletionTokensDetails(accepted_prediction_tokens=0, audio_tokens=0, reasoning_tokens=0, rejected_prediction_tokens=0), prompt_tokens_details=PromptTokensDetails(audio_tokens=0, cached_tokens=0))",
            "total_tokens": 3724,
        },
        "response2": "\"For breakfast, the choice is yours. But remember, 'A Jedi must have the deepest commitment, the most serious mind.' Start your day with something that gives you the energy of a podracer and the calm of a Jedi Knight meditating. Perhaps eggs with a side of fruits from Dagobah, or blue milk and a protein bantha steak. This is the Way to a great start!\"",
        "usage2": {
            "input_tokens": 3702,
            "output_tokens": 81,
            "cache_read_tokens": 0,
            "cache_write_tokens": 0,
            "reasoning_tokens": 0,
            "provider_tool_usage": None,
            "raw": "CompletionUsage(completion_tokens=81, prompt_tokens=3702, total_tokens=3783, completion_tokens_details=CompletionTokensDetails(accepted_prediction_tokens=0, audio_tokens=0, reasoning_tokens=0, rejected_prediction_tokens=0), prompt_tokens_details=PromptTokensDetails(audio_tokens=0, cached_tokens=0))",
            "total_tokens": 3783,
        },
    }
)



# This file was auto-generated by Fern from our API Definition.

import typing
from json.decoder import JSONDecodeError

from ..core.api_error import ApiError
from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.http_response import AsyncHttpResponse, HttpResponse
from ..core.jsonable_encoder import jsonable_encoder
from ..core.pydantic_utilities import parse_obj_as
from ..core.request_options import RequestOptions
from ..errors.bad_request_error import BadRequestError
from ..errors.conflict_error import ConflictError
from ..errors.forbidden_error import ForbiddenError
from ..errors.internal_server_error import InternalServerError
from ..errors.not_found_error import NotFoundError
from ..errors.payment_required_error import PaymentRequiredError
from ..errors.service_unavailable_error import ServiceUnavailableError
from ..errors.too_many_requests_error import TooManyRequestsError
from ..types.not_found_error_body import NotFoundErrorBody
from ..types.permission_denied_error import PermissionDeniedError
from ..types.plan_limit_exceeded_error import PlanLimitExceededError
from ..types.rate_limit_error import RateLimitError
from .types.organization_invitations_accept_response import (
    OrganizationInvitationsAcceptResponse,
)
from .types.organization_invitations_create_request_role import (
    OrganizationInvitationsCreateRequestRole,
)
from .types.organization_invitations_create_response import (
    OrganizationInvitationsCreateResponse,
)
from .types.organization_invitations_get_response import (
    OrganizationInvitationsGetResponse,
)
from .types.organization_invitations_list_response_item import (
    OrganizationInvitationsListResponseItem,
)

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class RawOrganizationInvitationsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def list(
        self,
        organization_id: str,
        *,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[typing.List[OrganizationInvitationsListResponseItem]]:
        """
        Parameters
        ----------
        organization_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[typing.List[OrganizationInvitationsListResponseItem]]
            Success
        """
        _response = self._client_wrapper.httpx_client.request(
            f"organizations/{jsonable_encoder(organization_id)}/invitations",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    typing.List[OrganizationInvitationsListResponseItem],
                    parse_obj_as(
                        type_=typing.List[OrganizationInvitationsListResponseItem],  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    def create(
        self,
        organization_id: str,
        *,
        recipient_email: str,
        role: OrganizationInvitationsCreateRequestRole,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[OrganizationInvitationsCreateResponse]:
        """
        Parameters
        ----------
        organization_id : str

        recipient_email : str
            a string matching the pattern ^[^ \t\n\r\f\v@]+@[^ \t\n\r\f\v@]+[.][^ \t\n\r\f\v@]+$

        role : OrganizationInvitationsCreateRequestRole

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[OrganizationInvitationsCreateResponse]
            Success
        """
        _response = self._client_wrapper.httpx_client.request(
            f"organizations/{jsonable_encoder(organization_id)}/invitations",
            method="POST",
            json={
                "recipientEmail": recipient_email,
                "role": role,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    OrganizationInvitationsCreateResponse,
                    parse_obj_as(
                        type_=OrganizationInvitationsCreateResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 402:
                raise PaymentRequiredError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PlanLimitExceededError,
                        parse_obj_as(
                            type_=PlanLimitExceededError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 409:
                raise ConflictError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    def get(
        self,
        organization_id: str,
        invitation_id: str,
        *,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[OrganizationInvitationsGetResponse]:
        """
        Parameters
        ----------
        organization_id : str

        invitation_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[OrganizationInvitationsGetResponse]
            Success
        """
        _response = self._client_wrapper.httpx_client.request(
            f"organizations/{jsonable_encoder(organization_id)}/invitations/{jsonable_encoder(invitation_id)}",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    OrganizationInvitationsGetResponse,
                    parse_obj_as(
                        type_=OrganizationInvitationsGetResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    def resend(
        self,
        organization_id: str,
        invitation_id: str,
        *,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Parameters
        ----------
        organization_id : str

        invitation_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            f"organizations/{jsonable_encoder(organization_id)}/invitations/{jsonable_encoder(invitation_id)}/resend",
            method="POST",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    def revoke(
        self,
        organization_id: str,
        invitation_id: str,
        *,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Parameters
        ----------
        organization_id : str

        invitation_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            f"organizations/{jsonable_encoder(organization_id)}/invitations/{jsonable_encoder(invitation_id)}/revoke",
            method="POST",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    def accept(
        self, *, token: str, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[OrganizationInvitationsAcceptResponse]:
        """
        Parameters
        ----------
        token : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[OrganizationInvitationsAcceptResponse]
            Success
        """
        _response = self._client_wrapper.httpx_client.request(
            "invitations/accept",
            method="POST",
            json={
                "token": token,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    OrganizationInvitationsAcceptResponse,
                    parse_obj_as(
                        type_=OrganizationInvitationsAcceptResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 402:
                raise PaymentRequiredError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PlanLimitExceededError,
                        parse_obj_as(
                            type_=PlanLimitExceededError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 409:
                raise ConflictError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )


class AsyncRawOrganizationInvitationsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def list(
        self,
        organization_id: str,
        *,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[typing.List[OrganizationInvitationsListResponseItem]]:
        """
        Parameters
        ----------
        organization_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[typing.List[OrganizationInvitationsListResponseItem]]
            Success
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"organizations/{jsonable_encoder(organization_id)}/invitations",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    typing.List[OrganizationInvitationsListResponseItem],
                    parse_obj_as(
                        type_=typing.List[OrganizationInvitationsListResponseItem],  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    async def create(
        self,
        organization_id: str,
        *,
        recipient_email: str,
        role: OrganizationInvitationsCreateRequestRole,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[OrganizationInvitationsCreateResponse]:
        """
        Parameters
        ----------
        organization_id : str

        recipient_email : str
            a string matching the pattern ^[^ \t\n\r\f\v@]+@[^ \t\n\r\f\v@]+[.][^ \t\n\r\f\v@]+$

        role : OrganizationInvitationsCreateRequestRole

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[OrganizationInvitationsCreateResponse]
            Success
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"organizations/{jsonable_encoder(organization_id)}/invitations",
            method="POST",
            json={
                "recipientEmail": recipient_email,
                "role": role,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    OrganizationInvitationsCreateResponse,
                    parse_obj_as(
                        type_=OrganizationInvitationsCreateResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 402:
                raise PaymentRequiredError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PlanLimitExceededError,
                        parse_obj_as(
                            type_=PlanLimitExceededError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 409:
                raise ConflictError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    async def get(
        self,
        organization_id: str,
        invitation_id: str,
        *,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[OrganizationInvitationsGetResponse]:
        """
        Parameters
        ----------
        organization_id : str

        invitation_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[OrganizationInvitationsGetResponse]
            Success
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"organizations/{jsonable_encoder(organization_id)}/invitations/{jsonable_encoder(invitation_id)}",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    OrganizationInvitationsGetResponse,
                    parse_obj_as(
                        type_=OrganizationInvitationsGetResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    async def resend(
        self,
        organization_id: str,
        invitation_id: str,
        *,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Parameters
        ----------
        organization_id : str

        invitation_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"organizations/{jsonable_encoder(organization_id)}/invitations/{jsonable_encoder(invitation_id)}/resend",
            method="POST",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    async def revoke(
        self,
        organization_id: str,
        invitation_id: str,
        *,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Parameters
        ----------
        organization_id : str

        invitation_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"organizations/{jsonable_encoder(organization_id)}/invitations/{jsonable_encoder(invitation_id)}/revoke",
            method="POST",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    async def accept(
        self, *, token: str, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[OrganizationInvitationsAcceptResponse]:
        """
        Parameters
        ----------
        token : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[OrganizationInvitationsAcceptResponse]
            Success
        """
        _response = await self._client_wrapper.httpx_client.request(
            "invitations/accept",
            method="POST",
            json={
                "token": token,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    OrganizationInvitationsAcceptResponse,
                    parse_obj_as(
                        type_=OrganizationInvitationsAcceptResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 402:
                raise PaymentRequiredError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PlanLimitExceededError,
                        parse_obj_as(
                            type_=PlanLimitExceededError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 409:
                raise ConflictError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )



from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    Text,
    ToolCall,
    ToolOutput,
    UserMessage,
)

test_snapshot = snapshot(
    {
        "response": {
            "provider_id": "google",
            "model_id": "google/gemini-3-pro-preview",
            "provider_model_name": "gemini-3-pro-preview",
            "finish_reason": None,
            "messages": [
                UserMessage(
                    content=[
                        Text(
                            text="Answer this question: What is 2 + 3, using the custom addition tool?"
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        ToolCall(
                            id="google_unknown_tool_id",
                            name="customAdditionImplementation",
                            args='{"a": 2, "b": 3}',
                        )
                    ],
                    provider_id="google",
                    model_id="google/gemini-3-pro-preview",
                    provider_model_name="gemini-3-pro-preview",
                    raw_message={
                        "parts": [
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": {
                                    "id": None,
                                    "args": {"a": 2, "b": 3},
                                    "name": "customAdditionImplementation",
                                    "partial_args": None,
                                    "will_continue": None,
                                },
                                "function_response": None,
                                "inline_data": None,
                                "text": None,
                                "thought": None,
                                "thought_signature": b"\x12\xde\x05\n\xdb\x05\x01r\xc8\xda|\x03\xd3T\xb2>\x03i\xc2\xbd\xb2\x14%O\xc3!]\xa9\xc8\xd0\xaec0\xa4\xd5j\xee7t\xbb\xa2\xca\xe9p\xa5Q\x81\x8cC\x12\xcd\xab\xbb,\x11\x8es%\x10\r\xe4Mq\x1a\x00\xd0\xd6\x8fQ!c\xac\x99\x0b\xab\xd0r\xa6\x00\x91\x8f\xab\xe3q&\x8fd\xae\xe0\xc2B/)/\xaf6;\xd8fZ\xc2\x95\x03\x0c\xd6\x06\xba\xbc\x83\xe6h\xfccN\x96w\x9e\xc1\xe3\xa5\xf2\xe6\xe3\x8f\xc3\xc8J9\x7f,\x16\xcd*\xc6c\xa2\xe1u\x80t\xbf\xcbX\x127 \xf2JW6\xef\x7f.\xb5~\xf7F\xaf\xdd\xbd\xc2?\x85\xf3s\x96\x04\xcc\xa1\xae\xea\x83v\xa9a\xec\xbb\x85_H\xff[_\xa9\x8aE\x8dC\x0b\x9ag\xe0\xf8\x8b\x11\xaaK]\xbc\xc0\xdb{\xf5\xef3N\xf2\x82\x07G \xc6\x92+\x1b\xd9+\xe2\xd9\xaf\x14w\x03\xde\x0b\xbc\xf9\x81\xfcc\xcc\xbd\x1d\xcc\xdc?\x18bT-\x96\x9d\x8fp\x8e\x86d\x8d\x1b\x9e\xc92\xed\xebx\te\x7f\x1d\x1cg\x93\n\xb1\xb5\x1eY\x17\xaaf9\\}\xa2\x8e7\xad\xbf&\x13\xac\xab\x0ey\xfa\xb3\xbe\xfaG\x05~\xe1^:h\xdc'\xec\xc7$\xcdC\xca[A|A{\x86\x02]\x7f\xee\x91c\xb5\xb1\xec\xdfym\xcb:\xe7\xde\xf3\x9e\x16\x1d\xb5bU\x10g\x15\x9e\x895\xb7\x1b\xbd\x8e\"\xe0\xf6\x06\xedS-D\x12\x96\x05\xe9\x1d\x8ayr\x9dL\xbf\xa1\x9f\x9f\x9a}\xcb\x94\xb0\xc1\x86\xfa\xb9\xce\xb2\xdbK*\xd6|D \x82\x82\x88\xe5\xb6*\xf2f\xb9\x9d(r\xc4\xe2\xdf\t\xf3\x0b\xeaLyq\xa9L\"\x05\xecZ2\x0b\x11\x9a\xb0\x9d\xc6\xe7\x9dL\x9a\x8b_X8ac'\xd9\x1d0\x8e\xd1J^\xa4q\xe8\xfc\x04\xe8\x04\xcc{\xa1\xa3\x03\x8c\xe3d@\xa7\xa1n\xe3+\xa45\t\x16\x8c\x9ed@\xe5\xae(U]\xb5a\x19\xf8/\x02\xddb\x8f\xe5\xde\xa8 \x05\x9c\x1b\xcc\xcd\xea\x82\x9e)I\xfalEO\x8e\xeb3\xce\xc4\x0b\xcc\xd9]\xcd.\xc4\xd8\xb9\x1c\xcd\xba\xd8#\xb0\x98\x16\x92@54\xa4\xe7d\nNApS\xe9\xec'.\xfa\xad&]\x11\t8FG\xbf\xb7j\xe8 2\xc4\xa9\x8f\x16L\xc0|\x0b\xa4a\x87\xf5Y\x9b\x8b\xc9\x8e\xc0\x10\xaak\x10)f\xbe\x87\xa15\xa2b\xe6B\x18\xd2M,P\x9bi\x10\x8a\xee\x1f@[\xb6H\x9fP\xb0\xcf\x87}\x0f:7c-q#\xbb6\xa3\xf0\x07\xfa\xdc3\xad(\xed\x92\xa5\x90\xde\xf2Kv\xcb\x13G\xf8\x91\x98\xee9\xee\xf2\xf6\x8d\xffr\x16\x9f]\xf7vP\xdb\x9f\xb3\x9e\x1e\xe4O=>\xc9\x90\x14\x1d\xe7\x0f\xba\xf5\xe0\xa830#`7\x91\xd7]r\x8e\xee~&H\xd0\xe4\x13'\x96\x1aA\x8a\xdc\xa0\xcdQ\xf2|kv\xf3\xb7b\xb5\x02\xe5\x13\x14\xb0zre\xc38\xc4\xe0\xa1:o\xceAP}\x9f\xb0\xf5\x9cz!~|a\xb0\xb6#l\xf5\xe6D\x1c,",
                                "video_metadata": None,
                            },
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": "",
                                "thought": None,
                                "thought_signature": None,
                                "video_metadata": None,
                            },
                        ],
                        "role": "model",
                    },
                ),
                UserMessage(
                    content=[
                        ToolOutput(
                            id="google_unknown_tool_id",
                            name="customAdditionImplementation",
                            result=13,
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
{
  "integer_a": 2,
  "integer_b": 3,
  "answer": 13
} \
"""
                        )
                    ],
                    provider_id="google",
                    model_id="google/gemini-3-pro-preview",
                    provider_model_name="gemini-3-pro-preview",
                    raw_message={
                        "parts": [
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": "{",
                                "thought": None,
                                "thought_signature": None,
                                "video_metadata": None,
                            },
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": """\

  "integer_a": 2,
  "integer_b\
""",
                                "thought": None,
                                "thought_signature": None,
                                "video_metadata": None,
                            },
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": """\
": 3,
  "answer": 13
} \
""",
                                "thought": None,
                                "thought_signature": None,
                                "video_metadata": None,
                            },
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": "",
                                "thought": None,
                                "thought_signature": None,
                                "video_metadata": None,
                            },
                        ],
                        "role": "model",
                    },
                ),
            ],
            "format": {
                "name": "IntegerAdditionResponse",
                "description": "A simple response for testing.",
                "schema": {
                    "description": "A simple response for testing.",
                    "properties": {
                        "integer_a": {"title": "Integer A", "type": "integer"},
                        "integer_b": {"title": "Integer B", "type": "integer"},
                        "answer": {"title": "Answer", "type": "integer"},
                    },
                    "required": ["integer_a", "integer_b", "answer"],
                    "title": "IntegerAdditionResponse",
                    "type": "object",
                },
                "mode": "strict",
                "formatting_instructions": None,
            },
            "tools": [
                {
                    "name": "customAdditionImplementation",
                    "description": "Perform computation on integers.",
                    "parameters": """\
{
  "properties": {
    "a": {
      "title": "A",
      "type": "integer"
    },
    "b": {
      "title": "B",
      "type": "integer"
    }
  },
  "required": [
    "a",
    "b"
  ],
  "additionalProperties": false,
  "defs": null
}\
""",
                    "strict": None,
                }
            ],
            "usage": {
                "input_tokens": 271,
                "output_tokens": 32,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 0,
                "provider_tool_usage": None,
                "raw": "None",
                "total_tokens": 303,
            },
            "n_chunks": 5,
        }
    }
)



"""LLM abstractions that aren't obstructions.

This module provides a unified interface for interacting with different LLM providers,
including messages, tools, response formatting, and streaming. It allows you to write
code that works with multiple LLM providers without changing your application logic.
"""

# TODO: Across the API, audit docstrings to ensure they are compliant Google-style docstrings
# (Write some tooling to ensure this happens consistently + in CI)

from contextlib import suppress

from . import (
    calls,
    content,
    exceptions,
    formatting,
    messages,
    models,
    prompts,
    providers,
    responses,
    retries,
    tools,
    types,
)

with suppress(ImportError):
    from . import mcp
from .calls import AsyncCall, AsyncContextCall, Call, CallDecorator, ContextCall, call
from .content import (
    AssistantContentChunk,
    AssistantContentPart,
    Audio,
    Base64AudioSource,
    Base64ImageSource,
    Document,
    Image,
    Text,
    TextChunk,
    TextEndChunk,
    TextStartChunk,
    Thought,
    ThoughtChunk,
    ThoughtEndChunk,
    ThoughtStartChunk,
    ToolCall,
    ToolCallChunk,
    ToolCallEndChunk,
    ToolCallStartChunk,
    ToolOutput,
    URLImageSource,
    UserContentPart,
)
from .context import Context, DepsT
from .exceptions import (
    APIError,
    AuthenticationError,
    BadRequestError,
    ConnectionError,
    Error,
    FeatureNotSupportedError,
    MissingAPIKeyError,
    NoRegisteredProviderError,
    NotFoundError,
    ParseError,
    PermissionError,
    ProviderError,
    RateLimitError,
    ResponseValidationError,
    RetriesExhausted,
    ServerError,
    StreamRestarted,
    TimeoutError,
    ToolError,
    ToolExecutionError,
    ToolNotFoundError,
)
from .formatting import (
    Format,
    FormatSpec,
    FormattableT,
    FormattingMode,
    OutputParser,
    Partial,
    format,
    output_parser,
)
from .messages import (
    AssistantContent,
    AssistantMessage,
    Message,
    SystemContent,
    SystemMessage,
    UserContent,
    UserMessage,
)
from .models import (
    Model,
    Params,
    ThinkingConfig,
    ThinkingLevel,
    model,
    model_from_context,
    use_model,
)
from .prompts import (
    AsyncContextPrompt,
    AsyncPrompt,
    ContextPrompt,
    Prompt,
    PromptDecorator,
    prompt,
)
from .providers import (
    ModelId,
    Provider,
    ProviderId,
    register_provider,
    reset_provider_registry,
)
from .responses import (
    AnyResponse,
    AsyncChunkIterator,
    AsyncContextResponse,
    AsyncContextStreamResponse,
    AsyncResponse,
    AsyncStream,
    AsyncStreamResponse,
    AsyncTextStream,
    AsyncThoughtStream,
    AsyncToolCallStream,
    ChunkIterator,
    ContextResponse,
    ContextStreamResponse,
    FinishReason,
    ProviderToolUsage,
    RawMessageChunk,
    Response,
    RootResponse,
    Stream,
    StreamResponse,
    StreamResponseChunk,
    TextStream,
    ThoughtStream,
    ToolCallStream,
    Usage,
    UsageDeltaChunk,
)
from .retries import (
    AsyncContextRetryResponse,
    AsyncContextRetryStreamResponse,
    AsyncRetryCall,
    AsyncRetryPrompt,
    AsyncRetryResponse,
    AsyncRetryStreamResponse,
    ContextRetryResponse,
    ContextRetryStreamResponse,
    RetryCall,
    RetryConfig,
    RetryModel,
    RetryModelParams,
    RetryPrompt,
    RetryResponse,
    RetryStreamResponse,
    retry,
    retry_model,
)
from .tools import (
    AnyToolFn,
    AnyTools,
    AnyToolSchema,
    AsyncContextTool,
    AsyncContextToolkit,
    AsyncContextTools,
    AsyncTool,
    AsyncToolkit,
    AsyncTools,
    BaseToolkit,
    ContextTool,
    ContextToolkit,
    ContextTools,
    ProviderTool,
    Tool,
    Toolkit,
    ToolkitT,
    Tools,
    ToolSchema,
    WebSearchTool,
    tool,
)
from .types import Jsonable

__all__ = [
    "APIError",
    "AnyResponse",
    "AnyToolFn",
    "AnyToolSchema",
    "AnyTools",
    "AssistantContent",
    "AssistantContentChunk",
    "AssistantContentPart",
    "AssistantMessage",
    "AsyncCall",
    "AsyncChunkIterator",
    "AsyncContextCall",
    "AsyncContextPrompt",
    "AsyncContextResponse",
    "AsyncContextRetryResponse",
    "AsyncContextRetryStreamResponse",
    "AsyncContextStreamResponse",
    "AsyncContextTool",
    "AsyncContextToolkit",
    "AsyncContextTools",
    "AsyncPrompt",
    "AsyncResponse",
    "AsyncRetryCall",
    "AsyncRetryPrompt",
    "AsyncRetryResponse",
    "AsyncRetryStreamResponse",
    "AsyncStream",
    "AsyncStreamResponse",
    "AsyncTextStream",
    "AsyncThoughtStream",
    "AsyncTool",
    "AsyncToolCallStream",
    "AsyncToolkit",
    "AsyncTools",
    "Audio",
    "AuthenticationError",
    "BadRequestError",
    "Base64AudioSource",
    "Base64ImageSource",
    "BaseToolkit",
    "Call",
    "CallDecorator",
    "ChunkIterator",
    "ConnectionError",
    "Context",
    "ContextCall",
    "ContextPrompt",
    "ContextResponse",
    "ContextRetryResponse",
    "ContextRetryStreamResponse",
    "ContextStreamResponse",
    "ContextTool",
    "ContextToolkit",
    "ContextTools",
    "DepsT",
    "Document",
    "Error",
    "FeatureNotSupportedError",
    "FinishReason",
    "Format",
    "FormatSpec",
    "FormattableT",
    "FormattingMode",
    "Image",
    "Jsonable",
    "Message",
    "MissingAPIKeyError",
    "Model",
    "ModelId",
    "NoRegisteredProviderError",
    "NotFoundError",
    "OutputParser",
    "Params",
    "ParseError",
    "Partial",
    "PermissionError",
    "Prompt",
    "PromptDecorator",
    "Provider",
    "ProviderError",
    "ProviderId",
    "ProviderTool",
    "ProviderToolUsage",
    "RateLimitError",
    "RawMessageChunk",
    "Response",
    "ResponseValidationError",
    "RetriesExhausted",
    "RetryCall",
    "RetryConfig",
    "RetryModel",
    "RetryModelParams",
    "RetryPrompt",
    "RetryResponse",
    "RetryStreamResponse",
    "RootResponse",
    "ServerError",
    "Stream",
    "StreamResponse",
    "StreamResponseChunk",
    "StreamRestarted",
    "SystemContent",
    "SystemMessage",
    "Text",
    "TextChunk",
    "TextEndChunk",
    "TextStartChunk",
    "TextStream",
    "ThinkingConfig",
    "ThinkingLevel",
    "Thought",
    "ThoughtChunk",
    "ThoughtEndChunk",
    "ThoughtStartChunk",
    "ThoughtStream",
    "TimeoutError",
    "Tool",
    "ToolCall",
    "ToolCallChunk",
    "ToolCallEndChunk",
    "ToolCallStartChunk",
    "ToolCallStream",
    "ToolError",
    "ToolExecutionError",
    "ToolNotFoundError",
    "ToolOutput",
    "ToolSchema",
    "Toolkit",
    "ToolkitT",
    "Tools",
    "URLImageSource",
    "Usage",
    "UsageDeltaChunk",
    "UserContent",
    "UserContentPart",
    "UserMessage",
    "WebSearchTool",
    "WebSearchTool",
    "call",
    "calls",
    "content",
    "exceptions",
    "format",
    "formatting",
    "mcp",
    "messages",
    "model",
    "model_from_context",
    "models",
    "output_parser",
    "prompt",
    "prompts",
    "providers",
    "register_provider",
    "reset_provider_registry",
    "responses",
    "retries",
    "retry",
    "retry_model",
    "tool",
    "tools",
    "types",
    "use_model",
]



"""RetryCall extends Call to return RetryResponse instead of Response."""

from typing import Generic, TypeVar, overload

from ..calls.calls import BaseCall
from ..formatting import FormattableT
from ..types import P
from .retry_config import RetryConfig
from .retry_models import RetryModel
from .retry_prompts import (
    AsyncRetryPrompt,
    BaseRetryPrompt,
    RetryPrompt,
)
from .retry_responses import (
    AsyncRetryResponse,
    RetryResponse,
)
from .retry_stream_responses import (
    AsyncRetryStreamResponse,
    RetryStreamResponse,
)

RetryPromptT = TypeVar("RetryPromptT", bound=BaseRetryPrompt)


class BaseRetryCall(BaseCall[RetryPromptT], Generic[RetryPromptT]):
    """Base class for retry-enabled calls with shared model wrapping."""

    @property
    def retry_config(self) -> RetryConfig:
        """The retry configuration from the prompt."""
        return self.prompt.retry_config

    @property
    def model(self) -> RetryModel:
        """The model used for generating responses. May be overwritten via `with llm.model(...)`."""
        model = super().model
        if isinstance(model, RetryModel):
            return model
        return RetryModel(model, self.retry_config)


class RetryCall(BaseRetryCall[RetryPrompt[P, FormattableT]], Generic[P, FormattableT]):
    """A retry-enabled call that extends BaseRetryCall and returns RetryResponse.

    This extends BaseRetryCall with a RetryPrompt and returns RetryResponse
    instead of Response. The prompt contains the retry configuration.

    The model can be overridden at runtime using `with llm.model(...)` context manager.
    """

    @overload
    def __call__(
        self: "RetryCall[P, None]", *args: P.args, **kwargs: P.kwargs
    ) -> RetryResponse[None]: ...

    @overload
    def __call__(
        self: "RetryCall[P, FormattableT]", *args: P.args, **kwargs: P.kwargs
    ) -> RetryResponse[FormattableT]: ...

    def __call__(
        self, *args: P.args, **kwargs: P.kwargs
    ) -> RetryResponse[None] | RetryResponse[FormattableT]:
        """Generates a retry response using the LLM."""
        return self.call(*args, **kwargs)

    @overload
    def call(
        self: "RetryCall[P, None]", *args: P.args, **kwargs: P.kwargs
    ) -> RetryResponse[None]: ...

    @overload
    def call(
        self: "RetryCall[P, FormattableT]", *args: P.args, **kwargs: P.kwargs
    ) -> RetryResponse[FormattableT]: ...

    def call(
        self, *args: P.args, **kwargs: P.kwargs
    ) -> RetryResponse[None] | RetryResponse[FormattableT]:
        """Generates a retry response using the LLM."""
        return self.prompt.call(self.model, *args, **kwargs)

    @overload
    def stream(
        self: "RetryCall[P, None]", *args: P.args, **kwargs: P.kwargs
    ) -> RetryStreamResponse[None]: ...

    @overload
    def stream(
        self: "RetryCall[P, FormattableT]", *args: P.args, **kwargs: P.kwargs
    ) -> RetryStreamResponse[FormattableT]: ...

    def stream(
        self, *args: P.args, **kwargs: P.kwargs
    ) -> RetryStreamResponse[None] | RetryStreamResponse[FormattableT]:
        """Generates a retry stream response using the LLM."""
        return self.prompt.stream(self.model, *args, **kwargs)


class AsyncRetryCall(
    BaseRetryCall[AsyncRetryPrompt[P, FormattableT]], Generic[P, FormattableT]
):
    """An async retry-enabled call that extends BaseRetryCall and returns AsyncRetryResponse.

    This extends BaseRetryCall with an AsyncRetryPrompt and returns AsyncRetryResponse
    instead of AsyncResponse. The prompt contains the retry configuration.

    The model can be overridden at runtime using `with llm.model(...)` context manager.
    """

    @overload
    async def __call__(
        self: "AsyncRetryCall[P, None]", *args: P.args, **kwargs: P.kwargs
    ) -> AsyncRetryResponse[None]: ...

    @overload
    async def __call__(
        self: "AsyncRetryCall[P, FormattableT]", *args: P.args, **kwargs: P.kwargs
    ) -> AsyncRetryResponse[FormattableT]: ...

    async def __call__(
        self, *args: P.args, **kwargs: P.kwargs
    ) -> AsyncRetryResponse[None] | AsyncRetryResponse[FormattableT]:
        """Generates a retry response using the LLM asynchronously."""
        return await self.call(*args, **kwargs)

    @overload
    async def call(
        self: "AsyncRetryCall[P, None]", *args: P.args, **kwargs: P.kwargs
    ) -> AsyncRetryResponse[None]: ...

    @overload
    async def call(
        self: "AsyncRetryCall[P, FormattableT]", *args: P.args, **kwargs: P.kwargs
    ) -> AsyncRetryResponse[FormattableT]: ...

    async def call(
        self, *args: P.args, **kwargs: P.kwargs
    ) -> AsyncRetryResponse[None] | AsyncRetryResponse[FormattableT]:
        """Generates a retry response using the LLM asynchronously."""
        return await self.prompt.call(self.model, *args, **kwargs)

    @overload
    async def stream(
        self: "AsyncRetryCall[P, None]", *args: P.args, **kwargs: P.kwargs
    ) -> AsyncRetryStreamResponse[None]: ...

    @overload
    async def stream(
        self: "AsyncRetryCall[P, FormattableT]", *args: P.args, **kwargs: P.kwargs
    ) -> AsyncRetryStreamResponse[FormattableT]: ...

    async def stream(
        self, *args: P.args, **kwargs: P.kwargs
    ) -> AsyncRetryStreamResponse[None] | AsyncRetryStreamResponse[FormattableT]:
        """Generates a retry stream response using the LLM asynchronously."""
        return await self.prompt.stream(self.model, *args, **kwargs)



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
import typing_extensions
from ...core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from ...core.serialization import FieldMetadata
from ...types.date import Date
from .organizations_preview_subscription_change_response_validation_errors_item import (
    OrganizationsPreviewSubscriptionChangeResponseValidationErrorsItem,
)


class OrganizationsPreviewSubscriptionChangeResponse(UniversalBaseModel):
    is_upgrade: typing_extensions.Annotated[bool, FieldMetadata(alias="isUpgrade")]
    prorated_amount_in_dollars: typing_extensions.Annotated[
        float, FieldMetadata(alias="proratedAmountInDollars")
    ]
    next_billing_date: typing_extensions.Annotated[
        Date, FieldMetadata(alias="nextBillingDate")
    ]
    recurring_amount_in_dollars: typing_extensions.Annotated[
        float, FieldMetadata(alias="recurringAmountInDollars")
    ]
    can_downgrade: typing_extensions.Annotated[
        typing.Optional[bool], FieldMetadata(alias="canDowngrade")
    ] = None
    validation_errors: typing_extensions.Annotated[
        typing.Optional[
            typing.List[
                OrganizationsPreviewSubscriptionChangeResponseValidationErrorsItem
            ]
        ],
        FieldMetadata(alias="validationErrors"),
    ] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(
            extra="allow", frozen=True
        )  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    Text,
    UserMessage,
)

sync_snapshot = snapshot(
    {
        "response": {
            "provider_id": "google",
            "model_id": "google/gemini-2.5-flash",
            "provider_model_name": "gemini-2.5-flash",
            "params": {"thinking": {"level": "medium", "include_thoughts": False}},
            "finish_reason": None,
            "usage": {
                "input_tokens": 30,
                "output_tokens": 2742,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 1973,
                "provider_tool_usage": None,
                "raw": """\
cache_tokens_details=None cached_content_token_count=None candidates_token_count=769 candidates_tokens_details=None prompt_token_count=30 prompt_tokens_details=[ModalityTokenCount(
  modality=<MediaModality.TEXT: 'TEXT'>,
  token_count=30
)] thoughts_token_count=1973 tool_use_prompt_token_count=None tool_use_prompt_tokens_details=None total_token_count=2772 traffic_type=None\
""",
                "total_tokens": 2772,
            },
            "messages": [
                UserMessage(
                    content=[
                        Text(
                            text="How many primes below 400 contain 79 as a substring? Answer ONLY with the number, not sharing which primes they are."
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
To find the number of primes below 400 that contain "79" as a substring, we can systematically check numbers in ranges.

1.  **Numbers less than 79:**
    No prime in this range can contain "79" as a substring, as "79" is a two-digit number.

2.  **The number 79 itself:**
    79 is a prime number, and it contains "79" as a substring.
    *Count = 1*

3.  **Numbers between 80 and 99:**
    No prime in this range can contain "79" as a substring, as the first digit would have to be 7.

4.  **Numbers between 100 and 399:**
    For a number `N` in this range to contain "79" as a substring, `N` must be a 3-digit number. The "79" substring can appear in two ways:
    *   **The number starts with 79 (e.g., 79X):**
        These numbers would be 790, 791, 792, ..., 799. All of these are greater than 400. So, no primes from this form are below 400.
    *   **The number ends with 79 (e.g., X79):**
        We need to check numbers of the form 179, 279, 379 for primality.
        *   **179:**
            To check if 179 is prime, we test divisibility by primes up to sqrt(179) which is approx 13.3.
            179 is not divisible by 2, 3 (1+7+9=17), 5.
            179 / 7 = 25 R 4
            179 / 11 = 16 R 3
            179 / 13 = 13 R 10
            So, 179 is prime.
            *Count = 1 + 1 = 2*
        *   **279:**
            The sum of digits is 2+7+9 = 18, which is divisible by 3.
            279 = 3 * 93 = 3 * 3 * 31. So, 279 is not prime.
        *   **379:**
            To check if 379 is prime, we test divisibility by primes up to sqrt(379) which is approx 19.4.
            379 is not divisible by 2, 3 (3+7+9=19), 5.
            379 / 7 = 54 R 1
            379 / 11 = 34 R 5
            379 / 13 = 29 R 2
            379 / 17 = 22 R 5
            379 / 19 = 19 R 18
            So, 379 is prime.
            *Count = 2 + 1 = 3*

The primes below 400 that contain "79" as a substring are 79, 179, and 379. There are 3 such primes.

3\
"""
                        )
                    ],
                    provider_id="google",
                    model_id="google/gemini-2.5-flash",
                    provider_model_name="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": """\
To find the number of primes below 400 that contain "79" as a substring, we can systematically check numbers in ranges.

1.  **Numbers less than 79:**
    No prime in this range can contain "79" as a substring, as "79" is a two-digit number.

2.  **The number 79 itself:**
    79 is a prime number, and it contains "79" as a substring.
    *Count = 1*

3.  **Numbers between 80 and 99:**
    No prime in this range can contain "79" as a substring, as the first digit would have to be 7.

4.  **Numbers between 100 and 399:**
    For a number `N` in this range to contain "79" as a substring, `N` must be a 3-digit number. The "79" substring can appear in two ways:
    *   **The number starts with 79 (e.g., 79X):**
        These numbers would be 790, 791, 792, ..., 799. All of these are greater than 400. So, no primes from this form are below 400.
    *   **The number ends with 79 (e.g., X79):**
        We need to check numbers of the form 179, 279, 379 for primality.
        *   **179:**
            To check if 179 is prime, we test divisibility by primes up to sqrt(179) which is approx 13.3.
            179 is not divisible by 2, 3 (1+7+9=17), 5.
            179 / 7 = 25 R 4
            179 / 11 = 16 R 3
            179 / 13 = 13 R 10
            So, 179 is prime.
            *Count = 1 + 1 = 2*
        *   **279:**
            The sum of digits is 2+7+9 = 18, which is divisible by 3.
            279 = 3 * 93 = 3 * 3 * 31. So, 279 is not prime.
        *   **379:**
            To check if 379 is prime, we test divisibility by primes up to sqrt(379) which is approx 19.4.
            379 is not divisible by 2, 3 (3+7+9=19), 5.
            379 / 7 = 54 R 1
            379 / 11 = 34 R 5
            379 / 13 = 29 R 2
            379 / 17 = 22 R 5
            379 / 19 = 19 R 18
            So, 379 is prime.
            *Count = 2 + 1 = 3*

The primes below 400 that contain "79" as a substring are 79, 179, and 379. There are 3 such primes.

3\
""",
                                "thought": None,
                                "thought_signature": None,
                                "video_metadata": None,
                            }
                        ],
                        "role": "model",
                    },
                ),
            ],
            "format": None,
            "tools": [],
        }
    }
)
async_snapshot = snapshot(
    {
        "response": {
            "provider_id": "google",
            "model_id": "google/gemini-2.5-flash",
            "provider_model_name": "gemini-2.5-flash",
            "params": {"thinking": {"level": "medium", "include_thoughts": False}},
            "finish_reason": None,
            "usage": {
                "input_tokens": 30,
                "output_tokens": 3069,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 3068,
                "provider_tool_usage": None,
                "raw": """\
cache_tokens_details=None cached_content_token_count=None candidates_token_count=1 candidates_tokens_details=None prompt_token_count=30 prompt_tokens_details=[ModalityTokenCount(
  modality=<MediaModality.TEXT: 'TEXT'>,
  token_count=30
)] thoughts_token_count=3068 tool_use_prompt_token_count=None tool_use_prompt_tokens_details=None total_token_count=3099 traffic_type=None\
""",
                "total_tokens": 3099,
            },
            "messages": [
                UserMessage(
                    content=[
                        Text(
                            text="How many primes below 400 contain 79 as a substring? Answer ONLY with the number, not sharing which primes they are."
                        )
                    ]
                ),
                AssistantMessage(
                    content=[Text(text="3")],
                    provider_id="google",
                    model_id="google/gemini-2.5-flash",
                    provider_model_name="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": "3",
                                "thought": None,
                                "thought_signature": None,
                                "video_metadata": None,
                            }
                        ],
                        "role": "model",
                    },
                ),
            ],
            "format": None,
            "tools": [],
        }
    }
)
stream_snapshot = snapshot(
    {
        "response": {
            "provider_id": "google",
            "model_id": "google/gemini-2.5-flash",
            "provider_model_name": "gemini-2.5-flash",
            "finish_reason": None,
            "messages": [
                UserMessage(
                    content=[
                        Text(
                            text="How many primes below 400 contain 79 as a substring? Answer ONLY with the number, not sharing which primes they are."
                        )
                    ]
                ),
                AssistantMessage(
                    content=[Text(text="3")],
                    provider_id="google",
                    model_id="google/gemini-2.5-flash",
                    provider_model_name="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": "3",
                                "thought": None,
                                "thought_signature": None,
                                "video_metadata": None,
                            }
                        ],
                        "role": "model",
                    },
                ),
            ],
            "format": None,
            "tools": [],
            "usage": {
                "input_tokens": 30,
                "output_tokens": 1,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 2814,
                "provider_tool_usage": None,
                "raw": "None",
                "total_tokens": 31,
            },
            "n_chunks": 3,
        }
    }
)
async_stream_snapshot = snapshot(
    {
        "response": {
            "provider_id": "google",
            "model_id": "google/gemini-2.5-flash",
            "provider_model_name": "gemini-2.5-flash",
            "finish_reason": None,
            "messages": [
                UserMessage(
                    content=[
                        Text(
                            text="How many primes below 400 contain 79 as a substring? Answer ONLY with the number, not sharing which primes they are."
                        )
                    ]
                ),
                AssistantMessage(
                    content=[Text(text="3")],
                    provider_id="google",
                    model_id="google/gemini-2.5-flash",
                    provider_model_name="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": "3",
                                "thought": None,
                                "thought_signature": None,
                                "video_metadata": None,
                            }
                        ],
                        "role": "model",
                    },
                ),
            ],
            "format": None,
            "tools": [],
            "usage": {
                "input_tokens": 30,
                "output_tokens": 1,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 2306,
                "provider_tool_usage": None,
                "raw": "None",
                "total_tokens": 31,
            },
            "n_chunks": 3,
        }
    }
)



"""OpenTelemetry integration tests for Response.resume instrumentation."""

from __future__ import annotations

from collections.abc import Generator

import pytest
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export.in_memory_span_exporter import InMemorySpanExporter

from mirascope import llm, ops


@pytest.fixture(autouse=True, scope="function")
def initialize(tracer_provider: TracerProvider) -> Generator[None, None, None]:
    """Initialize ops configuration and LLM instrumentation for each test."""
    ops.configure(tracer_provider=tracer_provider)
    ops.instrument_llm()
    yield
    ops.uninstrument_llm()


@pytest.mark.vcr()
def test_response_resume_exports_mirascope_span(
    span_exporter: InMemorySpanExporter,
) -> None:
    """Test Mirascope instrumentation for Response.resume."""

    @llm.tool
    def get_weather(location: str) -> str:
        """Get the current weather for a location."""
        return f"Weather in {location}: 72°F, sunny"

    model = llm.Model(model_id="openai/gpt-4o-mini")
    messages = [llm.messages.user("What's the weather in Paris?")]

    response = model.call(messages, tools=[get_weather])
    assert response.tool_calls

    tool_outputs = response.execute_tools()
    resumed = response.resume(tool_outputs)

    assert resumed.content

    spans = span_exporter.get_finished_spans()
    # 3 spans: Model.call, Response.resume (Mirascope), Model.resume (GenAI)
    assert len(spans) == 3

    # Mirascope span is at index 2 (outer span finishes after inner GenAI span)
    mirascope_span = spans[2]
    assert "Response.resume" in mirascope_span.name
    assert mirascope_span.attributes is not None
    assert mirascope_span.attributes.get("mirascope.type") == "response_resume"
    assert (
        mirascope_span.attributes.get("mirascope.response.model_id")
        == "openai/gpt-4o-mini"
    )
    assert mirascope_span.attributes.get("mirascope.response.provider_id") == "openai"
    assert "mirascope.trace.output" in mirascope_span.attributes
    # Verify the pretty() output is captured
    trace_output = mirascope_span.attributes.get("mirascope.trace.output")
    assert isinstance(trace_output, str)
    assert "assistant" in trace_output.lower() or "weather" in trace_output.lower()


@pytest.mark.vcr()
@pytest.mark.asyncio
async def test_async_response_resume_exports_mirascope_span(
    span_exporter: InMemorySpanExporter,
) -> None:
    """Test Mirascope instrumentation for AsyncResponse.resume."""

    @llm.tool
    async def get_weather(location: str) -> str:
        """Get the current weather for a location."""
        return f"Weather in {location}: 72°F, sunny"

    model = llm.Model(model_id="openai/gpt-4o-mini")
    messages = [llm.messages.user("What's the weather in Rome?")]

    response = await model.call_async(messages, tools=[get_weather])
    assert response.tool_calls

    tool_outputs = await response.execute_tools()
    resumed = await response.resume(tool_outputs)

    assert resumed.content

    spans = span_exporter.get_finished_spans()
    # 3 spans: Model.call_async, AsyncResponse.resume (Mirascope), Model.resume_async (GenAI)
    assert len(spans) == 3

    # Mirascope span is at index 2 (outer span finishes after inner GenAI span)
    mirascope_span = spans[2]
    assert "AsyncResponse.resume" in mirascope_span.name
    assert mirascope_span.attributes is not None
    assert mirascope_span.attributes.get("mirascope.type") == "response_resume"
    assert (
        mirascope_span.attributes.get("mirascope.response.model_id")
        == "openai/gpt-4o-mini"
    )
    assert mirascope_span.attributes.get("mirascope.response.provider_id") == "openai"
    assert "mirascope.trace.output" in mirascope_span.attributes


@pytest.mark.vcr()
def test_stream_response_resume_exports_mirascope_span(
    span_exporter: InMemorySpanExporter,
) -> None:
    """Test Mirascope instrumentation for StreamResponse.resume."""

    @llm.tool
    def get_weather(location: str) -> str:
        """Get the current weather for a location."""
        return f"Weather in {location}: 72°F, sunny"

    model = llm.Model(model_id="openai/gpt-4o-mini")
    messages = [llm.messages.user("What's the weather in Madrid?")]

    response = model.stream(messages, tools=[get_weather])
    response.finish()
    assert response.tool_calls

    tool_outputs = response.execute_tools()
    resumed = response.resume(tool_outputs)
    resumed.finish()

    assert resumed.content

    spans = span_exporter.get_finished_spans()
    # 3 spans: Model.stream, StreamResponse.resume (Mirascope), Model.resume_stream (GenAI)
    assert len(spans) == 3

    # For streaming, Mirascope span finishes first (at index 1), GenAI finishes after stream consumed (at index 2)
    mirascope_span = spans[1]
    assert "StreamResponse.resume" in mirascope_span.name
    assert mirascope_span.attributes is not None
    assert mirascope_span.attributes.get("mirascope.type") == "response_resume"
    assert (
        mirascope_span.attributes.get("mirascope.response.model_id")
        == "openai/gpt-4o-mini"
    )
    assert mirascope_span.attributes.get("mirascope.response.provider_id") == "openai"
    assert "mirascope.trace.output" in mirascope_span.attributes


@pytest.mark.vcr()
def test_context_response_resume_exports_mirascope_span(
    span_exporter: InMemorySpanExporter,
) -> None:
    """Test Mirascope instrumentation for ContextResponse.resume."""

    @llm.tool
    def get_weather(location: str) -> str:
        """Get the current weather for a location."""
        return f"Weather in {location}: 72°F, sunny"

    model = llm.Model(model_id="openai/gpt-4o-mini")
    messages = [llm.messages.user("What's the weather in Vienna?")]
    ctx = llm.Context(deps=None)

    response = model.context_call(messages, ctx=ctx, tools=[get_weather])
    assert response.tool_calls

    tool_outputs = response.execute_tools(ctx)
    resumed = response.resume(ctx, tool_outputs)

    assert resumed.content

    spans = span_exporter.get_finished_spans()
    # 3 spans: Model.context_call, ContextResponse.resume (Mirascope), Model.context_resume (GenAI)
    assert len(spans) == 3

    # Mirascope span is at index 2 (outer span finishes after inner GenAI span)
    mirascope_span = spans[2]
    assert "ContextResponse.resume" in mirascope_span.name
    assert mirascope_span.attributes is not None
    assert mirascope_span.attributes.get("mirascope.type") == "response_resume"
    assert (
        mirascope_span.attributes.get("mirascope.response.model_id")
        == "openai/gpt-4o-mini"
    )
    assert mirascope_span.attributes.get("mirascope.response.provider_id") == "openai"
    assert "mirascope.trace.output" in mirascope_span.attributes


@pytest.mark.vcr()
@pytest.mark.asyncio
async def test_async_stream_response_resume_exports_mirascope_span(
    span_exporter: InMemorySpanExporter,
) -> None:
    """Test Mirascope instrumentation for AsyncStreamResponse.resume."""

    @llm.tool
    async def get_weather(location: str) -> str:
        """Get the current weather for a location."""
        return f"Weather in {location}: 72°F, sunny"

    model = llm.Model(model_id="openai/gpt-4o-mini")
    messages = [llm.messages.user("What's the weather in Berlin?")]

    response = await model.stream_async(messages, tools=[get_weather])
    await response.finish()
    assert response.tool_calls

    tool_outputs = await response.execute_tools()
    resumed = await response.resume(tool_outputs)
    await resumed.finish()

    assert resumed.content

    spans = span_exporter.get_finished_spans()
    # 3 spans: Model.stream_async, AsyncStreamResponse.resume (Mirascope), Model.resume_stream_async (GenAI)
    assert len(spans) == 3

    # For streaming, Mirascope span finishes first (at index 1), GenAI finishes after stream consumed (at index 2)
    mirascope_span = spans[1]
    assert "AsyncStreamResponse.resume" in mirascope_span.name
    assert mirascope_span.attributes is not None
    assert mirascope_span.attributes.get("mirascope.type") == "response_resume"
    assert (
        mirascope_span.attributes.get("mirascope.response.model_id")
        == "openai/gpt-4o-mini"
    )
    assert mirascope_span.attributes.get("mirascope.response.provider_id") == "openai"
    assert "mirascope.trace.output" in mirascope_span.attributes


@pytest.mark.vcr()
@pytest.mark.asyncio
async def test_async_context_response_resume_exports_mirascope_span(
    span_exporter: InMemorySpanExporter,
) -> None:
    """Test Mirascope instrumentation for AsyncContextResponse.resume."""

    @llm.tool
    async def get_weather(location: str) -> str:
        """Get the current weather for a location."""
        return f"Weather in {location}: 72°F, sunny"

    model = llm.Model(model_id="openai/gpt-4o-mini")
    messages = [llm.messages.user("What's the weather in Prague?")]
    ctx = llm.Context(deps=None)

    response = await model.context_call_async(messages, ctx=ctx, tools=[get_weather])
    assert response.tool_calls

    tool_outputs = await response.execute_tools(ctx)
    resumed = await response.resume(ctx, tool_outputs)

    assert resumed.content

    spans = span_exporter.get_finished_spans()
    # 3 spans: Model.context_call_async, AsyncContextResponse.resume (Mirascope), Model.context_resume_async (GenAI)
    assert len(spans) == 3

    # Mirascope span is at index 2 (outer span finishes after inner GenAI span)
    mirascope_span = spans[2]
    assert "AsyncContextResponse.resume" in mirascope_span.name
    assert mirascope_span.attributes is not None
    assert mirascope_span.attributes.get("mirascope.type") == "response_resume"
    assert (
        mirascope_span.attributes.get("mirascope.response.model_id")
        == "openai/gpt-4o-mini"
    )
    assert mirascope_span.attributes.get("mirascope.response.provider_id") == "openai"
    assert "mirascope.trace.output" in mirascope_span.attributes


@pytest.mark.vcr()
def test_context_stream_response_resume_exports_mirascope_span(
    span_exporter: InMemorySpanExporter,
) -> None:
    """Test Mirascope instrumentation for ContextStreamResponse.resume."""

    @llm.tool
    def get_weather(location: str) -> str:
        """Get the current weather for a location."""
        return f"Weather in {location}: 72°F, sunny"

    model = llm.Model(model_id="openai/gpt-4o-mini")
    messages = [llm.messages.user("What's the weather in Amsterdam?")]
    ctx = llm.Context(deps=None)

    response = model.context_stream(messages, ctx=ctx, tools=[get_weather])
    response.finish()
    assert response.tool_calls

    tool_outputs = response.execute_tools(ctx)
    resumed = response.resume(ctx, tool_outputs)
    resumed.finish()

    assert resumed.content

    spans = span_exporter.get_finished_spans()
    # 3 spans: Model.context_stream, ContextStreamResponse.resume (Mirascope), Model.context_resume_stream (GenAI)
    assert len(spans) == 3

    # For streaming, Mirascope span finishes first (at index 1), GenAI finishes after stream consumed (at index 2)
    mirascope_span = spans[1]
    assert "ContextStreamResponse.resume" in mirascope_span.name
    assert mirascope_span.attributes is not None
    assert mirascope_span.attributes.get("mirascope.type") == "response_resume"
    assert (
        mirascope_span.attributes.get("mirascope.response.model_id")
        == "openai/gpt-4o-mini"
    )
    assert mirascope_span.attributes.get("mirascope.response.provider_id") == "openai"
    assert "mirascope.trace.output" in mirascope_span.attributes


@pytest.mark.vcr()
@pytest.mark.asyncio
async def test_async_context_stream_response_resume_exports_mirascope_span(
    span_exporter: InMemorySpanExporter,
) -> None:
    """Test Mirascope instrumentation for AsyncContextStreamResponse.resume."""

    @llm.tool
    async def get_weather(location: str) -> str:
        """Get the current weather for a location."""
        return f"Weather in {location}: 72°F, sunny"

    model = llm.Model(model_id="openai/gpt-4o-mini")
    messages = [llm.messages.user("What's the weather in Brussels?")]
    ctx = llm.Context(deps=None)

    response = await model.context_stream_async(messages, ctx=ctx, tools=[get_weather])
    await response.finish()
    assert response.tool_calls

    tool_outputs = await response.execute_tools(ctx)
    resumed = await response.resume(ctx, tool_outputs)
    await resumed.finish()

    assert resumed.content

    spans = span_exporter.get_finished_spans()
    # 3 spans: Model.context_stream_async, AsyncContextStreamResponse.resume (Mirascope), Model.context_resume_stream_async (GenAI)
    assert len(spans) == 3

    # For streaming, Mirascope span finishes first (at index 1), GenAI finishes after stream consumed (at index 2)
    mirascope_span = spans[1]
    assert "AsyncContextStreamResponse.resume" in mirascope_span.name
    assert mirascope_span.attributes is not None
    assert mirascope_span.attributes.get("mirascope.type") == "response_resume"
    assert (
        mirascope_span.attributes.get("mirascope.response.model_id")
        == "openai/gpt-4o-mini"
    )
    assert mirascope_span.attributes.get("mirascope.response.provider_id") == "openai"
    assert "mirascope.trace.output" in mirascope_span.attributes
    assert "mirascope.trace.output" in mirascope_span.attributes



from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    SystemMessage,
    Text,
    UserMessage,
)

sync_snapshot = snapshot(
    {
        "response": {
            "provider_id": "google",
            "model_id": "google/gemini-2.5-flash",
            "provider_model_name": "gemini-2.5-flash",
            "params": {},
            "finish_reason": None,
            "usage": {
                "input_tokens": 215,
                "output_tokens": 82,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 70,
                "provider_tool_usage": None,
                "raw": """\
cache_tokens_details=None cached_content_token_count=None candidates_token_count=12 candidates_tokens_details=None prompt_token_count=215 prompt_tokens_details=[ModalityTokenCount(
  modality=<MediaModality.TEXT: 'TEXT'>,
  token_count=215
)] thoughts_token_count=70 tool_use_prompt_token_count=None tool_use_prompt_tokens_details=None total_token_count=297 traffic_type=None\
""",
                "total_tokens": 297,
            },
            "messages": [
                SystemMessage(
                    content=Text(
                        text="""\
Respond only with valid JSON that matches this exact schema:
{
  "description": "A lucky number chosen by the model.",
  "properties": {
    "value": {
      "title": "Value",
      "type": "integer"
    }
  },
  "required": [
    "value"
  ],
  "title": "LuckyNumber",
  "type": "object"
}\
"""
                    )
                ),
                UserMessage(
                    content=[Text(text="Pick a random number between 1 and 1000")]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
{
  "value": 731
}\
"""
                        )
                    ],
                    provider_id="google",
                    model_id="google/gemini-2.5-flash",
                    provider_model_name="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": """\
{
  "value": 731
}\
""",
                                "thought": None,
                                "thought_signature": None,
                                "video_metadata": None,
                            }
                        ],
                        "role": "model",
                    },
                ),
                UserMessage(
                    content=[
                        Text(
                            text="""\
Your response failed schema validation:
1 validation error for LuckyNumber
value
  Value error, The number must be exactly 173, but got 731 [type=value_error, input_value=731, input_type=int]
    For further information visit https://errors.pydantic.dev/2.12/v/value_error

Please correct these issues and respond again.\
"""
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
{
  "value": 173
}\
"""
                        )
                    ],
                    provider_id="google",
                    model_id="google/gemini-2.5-flash",
                    provider_model_name="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": """\
{
  "value": 173
}\
""",
                                "thought": None,
                                "thought_signature": None,
                                "video_metadata": None,
                            }
                        ],
                        "role": "model",
                    },
                ),
            ],
            "format": {
                "name": "LuckyNumber",
                "description": "A lucky number chosen by the model.",
                "schema": {
                    "description": "A lucky number chosen by the model.",
                    "properties": {"value": {"title": "Value", "type": "integer"}},
                    "required": ["value"],
                    "title": "LuckyNumber",
                    "type": "object",
                },
                "mode": "json",
                "formatting_instructions": """\
Respond only with valid JSON that matches this exact schema:
{
  "description": "A lucky number chosen by the model.",
  "properties": {
    "value": {
      "title": "Value",
      "type": "integer"
    }
  },
  "required": [
    "value"
  ],
  "title": "LuckyNumber",
  "type": "object"
}\
""",
            },
            "tools": [],
        }
    }
)
async_snapshot = snapshot(
    {
        "response": {
            "provider_id": "google",
            "model_id": "google/gemini-2.5-flash",
            "provider_model_name": "gemini-2.5-flash",
            "params": {},
            "finish_reason": None,
            "usage": {
                "input_tokens": 215,
                "output_tokens": 62,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 50,
                "provider_tool_usage": None,
                "raw": """\
cache_tokens_details=None cached_content_token_count=None candidates_token_count=12 candidates_tokens_details=None prompt_token_count=215 prompt_tokens_details=[ModalityTokenCount(
  modality=<MediaModality.TEXT: 'TEXT'>,
  token_count=215
)] thoughts_token_count=50 tool_use_prompt_token_count=None tool_use_prompt_tokens_details=None total_token_count=277 traffic_type=None\
""",
                "total_tokens": 277,
            },
            "messages": [
                SystemMessage(
                    content=Text(
                        text="""\
Respond only with valid JSON that matches this exact schema:
{
  "description": "A lucky number chosen by the model.",
  "properties": {
    "value": {
      "title": "Value",
      "type": "integer"
    }
  },
  "required": [
    "value"
  ],
  "title": "LuckyNumber",
  "type": "object"
}\
"""
                    )
                ),
                UserMessage(
                    content=[Text(text="Pick a random number between 1 and 1000")]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
{
  "value": 731
}\
"""
                        )
                    ],
                    provider_id="google",
                    model_id="google/gemini-2.5-flash",
                    provider_model_name="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": """\
{
  "value": 731
}\
""",
                                "thought": None,
                                "thought_signature": None,
                                "video_metadata": None,
                            }
                        ],
                        "role": "model",
                    },
                ),
                UserMessage(
                    content=[
                        Text(
                            text="""\
Your response failed schema validation:
1 validation error for LuckyNumber
value
  Value error, The number must be exactly 173, but got 731 [type=value_error, input_value=731, input_type=int]
    For further information visit https://errors.pydantic.dev/2.12/v/value_error

Please correct these issues and respond again.\
"""
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
{
  "value": 173
}\
"""
                        )
                    ],
                    provider_id="google",
                    model_id="google/gemini-2.5-flash",
                    provider_model_name="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": """\
{
  "value": 173
}\
""",
                                "thought": None,
                                "thought_signature": None,
                                "video_metadata": None,
                            }
                        ],
                        "role": "model",
                    },
                ),
            ],
            "format": {
                "name": "LuckyNumber",
                "description": "A lucky number chosen by the model.",
                "schema": {
                    "description": "A lucky number chosen by the model.",
                    "properties": {"value": {"title": "Value", "type": "integer"}},
                    "required": ["value"],
                    "title": "LuckyNumber",
                    "type": "object",
                },
                "mode": "json",
                "formatting_instructions": """\
Respond only with valid JSON that matches this exact schema:
{
  "description": "A lucky number chosen by the model.",
  "properties": {
    "value": {
      "title": "Value",
      "type": "integer"
    }
  },
  "required": [
    "value"
  ],
  "title": "LuckyNumber",
  "type": "object"
}\
""",
            },
            "tools": [],
        }
    }
)
stream_snapshot = snapshot(
    {
        "response": {
            "provider_id": "google",
            "model_id": "google/gemini-2.5-flash",
            "provider_model_name": "gemini-2.5-flash",
            "finish_reason": None,
            "messages": [
                SystemMessage(
                    content=Text(
                        text="""\
Respond only with valid JSON that matches this exact schema:
{
  "description": "A lucky number chosen by the model.",
  "properties": {
    "value": {
      "title": "Value",
      "type": "integer"
    }
  },
  "required": [
    "value"
  ],
  "title": "LuckyNumber",
  "type": "object"
}\
"""
                    )
                ),
                UserMessage(
                    content=[Text(text="Pick a random number between 1 and 1000")]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
{
  "value": 573
}\
"""
                        )
                    ],
                    provider_id="google",
                    model_id="google/gemini-2.5-flash",
                    provider_model_name="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": """\
{
  "value": 57\
""",
                                "thought": None,
                                "thought_signature": None,
                                "video_metadata": None,
                            },
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": """\
3
}\
""",
                                "thought": None,
                                "thought_signature": None,
                                "video_metadata": None,
                            },
                        ],
                        "role": "model",
                    },
                ),
                UserMessage(
                    content=[
                        Text(
                            text="""\
Your response failed schema validation:
1 validation error for LuckyNumber
value
  Value error, The number must be exactly 173, but got 573 [type=value_error, input_value=573, input_type=int]
    For further information visit https://errors.pydantic.dev/2.12/v/value_error

Please correct these issues and respond again.\
"""
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
{
  "value": 173
}\
"""
                        )
                    ],
                    provider_id="google",
                    model_id="google/gemini-2.5-flash",
                    provider_model_name="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": """\
{
  "value": 173
}\
""",
                                "thought": None,
                                "thought_signature": None,
                                "video_metadata": None,
                            }
                        ],
                        "role": "model",
                    },
                ),
            ],
            "format": {
                "name": "LuckyNumber",
                "description": "A lucky number chosen by the model.",
                "schema": {
                    "description": "A lucky number chosen by the model.",
                    "properties": {"value": {"title": "Value", "type": "integer"}},
                    "required": ["value"],
                    "title": "LuckyNumber",
                    "type": "object",
                },
                "mode": "json",
                "formatting_instructions": """\
Respond only with valid JSON that matches this exact schema:
{
  "description": "A lucky number chosen by the model.",
  "properties": {
    "value": {
      "title": "Value",
      "type": "integer"
    }
  },
  "required": [
    "value"
  ],
  "title": "LuckyNumber",
  "type": "object"
}\
""",
            },
            "tools": [],
            "usage": {
                "input_tokens": 215,
                "output_tokens": 12,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 84,
                "provider_tool_usage": None,
                "raw": "None",
                "total_tokens": 227,
            },
            "n_chunks": 3,
        }
    }
)
async_stream_snapshot = snapshot(
    {
        "response": {
            "provider_id": "google",
            "model_id": "google/gemini-2.5-flash",
            "provider_model_name": "gemini-2.5-flash",
            "finish_reason": None,
            "messages": [
                SystemMessage(
                    content=Text(
                        text="""\
Respond only with valid JSON that matches this exact schema:
{
  "description": "A lucky number chosen by the model.",
  "properties": {
    "value": {
      "title": "Value",
      "type": "integer"
    }
  },
  "required": [
    "value"
  ],
  "title": "LuckyNumber",
  "type": "object"
}\
"""
                    )
                ),
                UserMessage(
                    content=[Text(text="Pick a random number between 1 and 1000")]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
{
  "value": 731
}\
"""
                        )
                    ],
                    provider_id="google",
                    model_id="google/gemini-2.5-flash",
                    provider_model_name="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": """\
{
  "value": 7\
""",
                                "thought": None,
                                "thought_signature": None,
                                "video_metadata": None,
                            },
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": """\
31
}\
""",
                                "thought": None,
                                "thought_signature": None,
                                "video_metadata": None,
                            },
                        ],
                        "role": "model",
                    },
                ),
                UserMessage(
                    content=[
                        Text(
                            text="""\
Your response failed schema validation:
1 validation error for LuckyNumber
value
  Value error, The number must be exactly 173, but got 731 [type=value_error, input_value=731, input_type=int]
    For further information visit https://errors.pydantic.dev/2.12/v/value_error

Please correct these issues and respond again.\
"""
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
{
  "value": 173
}\
"""
                        )
                    ],
                    provider_id="google",
                    model_id="google/gemini-2.5-flash",
                    provider_model_name="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": """\
{
  "value": 173
}\
""",
                                "thought": None,
                                "thought_signature": None,
                                "video_metadata": None,
                            }
                        ],
                        "role": "model",
                    },
                ),
            ],
            "format": {
                "name": "LuckyNumber",
                "description": "A lucky number chosen by the model.",
                "schema": {
                    "description": "A lucky number chosen by the model.",
                    "properties": {"value": {"title": "Value", "type": "integer"}},
                    "required": ["value"],
                    "title": "LuckyNumber",
                    "type": "object",
                },
                "mode": "json",
                "formatting_instructions": """\
Respond only with valid JSON that matches this exact schema:
{
  "description": "A lucky number chosen by the model.",
  "properties": {
    "value": {
      "title": "Value",
      "type": "integer"
    }
  },
  "required": [
    "value"
  ],
  "title": "LuckyNumber",
  "type": "object"
}\
""",
            },
            "tools": [],
            "usage": {
                "input_tokens": 215,
                "output_tokens": 12,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 73,
                "provider_tool_usage": None,
                "raw": "None",
                "total_tokens": 227,
            },
            "n_chunks": 3,
        }
    }
)
